/*
Program: cse
Version: 1.0
Date: 15th November 2021
Author: Gerard Leclercq
E-mail: kotom.eng@free.fr

Copyright 2021 Gerard Leclercq

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License,
or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<math.h>

#define	ENTITY_MAX 1000000
#define	FILE_LINE_MAX_SIZE 1000
#define	NPLUS_MAX_DIGIT 16	// Max number of digits of n+
							// 2^32 = 4 294 967 296, hence 10 digits (< 16)

int main(argc, argv)
int 	argc;
char	*argv[];
{
	char	color[NPLUS_MAX_DIGIT];
	char	display[NPLUS_MAX_DIGIT];
	char	esn[NPLUS_MAX_DIGIT]; // Entity Sequence Number
	char	end_of_entity[8]; // End of Entity |~
	char	file_line[FILE_LINE_MAX_SIZE];
	char	line_type[NPLUS_MAX_DIGIT];
	char	line_width[NPLUS_MAX_DIGIT];
	double	angle1, angle2; // For ARC
	double	diameter, length; // Entity dimension
	double	max_length; // Entity dimension to be deleted
	static double	x[ENTITY_MAX], y[ENTITY_MAX]; // Coordinates of points.
	int		esn1, esns; // Entity Sequence Number
	int		i; // Counter
	int		n_property; // Number of property
	static int		property[ENTITY_MAX]; // Property
	int		c1, p1, p2; // Points
	int		n_ARC, n_ARC_deleted; // Number of ARC, total and deleted
	int		n_ARC_1, n_ARC_2, n_ARC_5; // Number of ARC of length 1, 2 and 5
	int		n_CIR, n_CIR_deleted; // Number of CIR, total and deleted
	int		n_CIR_1, n_CIR_2, n_CIR_5; // Number of CIR of diameter 1, 2 and 5
	int		n_LIN, n_LIN_deleted; // Number of LIN, total and deleted
	int		n_LIN_1, n_LIN_2, n_LIN_5; // Number of LIN of length 1, 2 and 5
	int		n_point; // Number of points of current section #~61
	int		n_point_max; // Biggest number of points of section #~61 of the file
	FILE	*fp_in, *fp_out, *fp_log;
	
	if (argc < 5) {
		printf("Usage: cse input_file output_file log_file size\n");
		exit(EXIT_FAILURE);
	}
	
	fp_in = fopen(argv[1], "r");
	fp_out = fopen(argv[2], "w");
	fp_log = fopen(argv[3], "w");
	max_length = atof(argv[4]);
	
	fprintf(fp_log, "Clean Small Elements in MI files\n");
	fprintf(fp_log, "Web site: http://kotom.eng.free.fr\n");
	fprintf(fp_log, "sizeof(int) : %2d\n", sizeof(int));
	
	n_LIN = 0;
	n_LIN_deleted = 0;
	n_LIN_1 = 0;
	n_LIN_2 = 0;
	n_LIN_5 = 0;
	
	n_ARC = 0;
	n_ARC_deleted = 0;
	n_ARC_1 = 0;
	n_ARC_2 = 0;
	n_ARC_5 = 0;
	
	n_CIR = 0;
	n_CIR_deleted = 0;
	n_CIR_1 = 0;
	n_CIR_2 = 0;
	n_CIR_5 = 0;
	
	n_point_max = 0;
	
	fgets(file_line, FILE_LINE_MAX_SIZE, fp_in);
	
	do {
		if (! strcmp(file_line, "#~3\n")){ // Setup section
			// Get Power for Largest Integer
			fprintf(fp_out, "%s", file_line);
			for (i = 0; i < 27; ++i) {
				fgets(file_line, FILE_LINE_MAX_SIZE, fp_in);
				fprintf(fp_out, "%s", file_line);
			}
			fprintf(fp_log, "Power for Largest Integer : %s", file_line);
		}
		
		else if (! strcmp(file_line, "#~61\n")){ // Point section
			fprintf(fp_out, "%s", file_line);
			fgets(file_line, FILE_LINE_MAX_SIZE, fp_in); // P or #~62 or #~
			fprintf(fp_out, "%s", file_line);
			
			// Get matrix offset esn1 with first point
			if (! strcmp(file_line, "P\n")) {
				n_point = 1;
				esn1 = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // First Entity Sequence Number of the section
				fprintf(fp_out, "%s", file_line);
				x[0] = atof(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // X Coordinates of Point
				fprintf(fp_out, "%s", file_line);
				y[0] = atof(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Y Coordinates of Point
				fprintf(fp_out, "%s", file_line);
				fgets(file_line, FILE_LINE_MAX_SIZE, fp_in); // |~
				fprintf(fp_out, "%s", file_line);
				fgets(file_line, FILE_LINE_MAX_SIZE, fp_in); // P or #~62
				fprintf(fp_out, "%s", file_line);
			}
			
			// Get other points
			while (strncmp(file_line, "#~", 2)) {
				++n_point;
				esns = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)) - esn1; // Entity Sequence Number relative to the section
				fprintf(fp_out, "%s", file_line);
				x[esns] = atof(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // X Coordinates of Point
				fprintf(fp_out, "%s", file_line);
				y[esns] = atof(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Y Coordinates of Point
				fprintf(fp_out, "%s", file_line);
				fgets(file_line, FILE_LINE_MAX_SIZE, fp_in); // |~
				fprintf(fp_out, "%s", file_line);
				fgets(file_line, FILE_LINE_MAX_SIZE, fp_in); // P or #~62
				fprintf(fp_out, "%s", file_line);
			}
			
			if (n_point > n_point_max)
				n_point_max = n_point;
		}
		
		else if (! strcmp(file_line, "LIN\n")){
			++n_LIN;
			fgets(esn, FILE_LINE_MAX_SIZE, fp_in); // Entity Sequence Number
			fgets(color, FILE_LINE_MAX_SIZE, fp_in);
			fgets(line_type, FILE_LINE_MAX_SIZE, fp_in);
			fgets(line_width, FILE_LINE_MAX_SIZE, fp_in);
			fgets(display, FILE_LINE_MAX_SIZE, fp_in);
			n_property = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Property Count
			for (i = 0; i < n_property; ++i)
				property[i] = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Property pointer
			p1 = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Pointer to Line Begin Point
			p2 = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Pointer to Line End Point
			fgets(end_of_entity, FILE_LINE_MAX_SIZE, fp_in); // |~
			
			length = hypot(x[p1 - esn1] - x[p2 - esn1] , y[p1 - esn1] - y[p2 - esn1]);
			
			if (length <= 1) ++n_LIN_1;
			if (length <= 2) ++n_LIN_2;
			if (length <= 5) ++n_LIN_5;
			
			if (length > max_length) {
				fprintf(fp_out, "%s", "LIN\n");
				fprintf(fp_out, "%s", esn);
				fprintf(fp_out, "%s", color);
				fprintf(fp_out, "%s", line_type);
				fprintf(fp_out, "%s", line_width);
				fprintf(fp_out, "%s", display);
				fprintf(fp_out, "%d\n", n_property);
				for (i = 0; i < n_property; ++i)
					fprintf(fp_out, "%d\n", property[i]);
				fprintf(fp_out, "%d\n", p1);
				fprintf(fp_out, "%d\n", p2);
				fprintf(fp_out, "%s", end_of_entity);
			}
			else
				++n_LIN_deleted;
		}
		
		else if (! strcmp(file_line, "CIR\n")){
			++n_CIR;
			fgets(esn, FILE_LINE_MAX_SIZE, fp_in); // Entity Sequence Number
			fgets(color, FILE_LINE_MAX_SIZE, fp_in);
			fgets(line_type, FILE_LINE_MAX_SIZE, fp_in);
			fgets(line_width, FILE_LINE_MAX_SIZE, fp_in);
			fgets(display, FILE_LINE_MAX_SIZE, fp_in);
			n_property = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Property Count
			for (i = 0; i < n_property; ++i)
				property[i] = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Property pointer
			p1 = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Pointer to Circle Center Point
			p2 = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Circle Circumference Point
			fgets(end_of_entity, FILE_LINE_MAX_SIZE, fp_in); // |~
			
			diameter = 2 * hypot(x[p1 - esn1] - x[p2 - esn1] , y[p1 - esn1] - y[p2 - esn1]);
			
			if (diameter <= 1) ++n_CIR_1;
			if (diameter <= 2) ++n_CIR_2;
			if (diameter <= 5) ++n_CIR_5;
			
			if (diameter > max_length) {
				fprintf(fp_out, "%s", "CIR\n");
				fprintf(fp_out, "%s", esn);
				fprintf(fp_out, "%s", color);
				fprintf(fp_out, "%s", line_type);
				fprintf(fp_out, "%s", line_width);
				fprintf(fp_out, "%s", display);
				fprintf(fp_out, "%d\n", n_property);
				for (i = 0; i < n_property; ++i)
					fprintf(fp_out, "%d\n", property[i]);
				fprintf(fp_out, "%d\n", p1);
				fprintf(fp_out, "%d\n", p2);
				fprintf(fp_out, "%s", end_of_entity);
			}
			else
				++n_CIR_deleted;
		}
		
		else if (! strcmp(file_line, "ARC\n")){
			++n_ARC;
			fgets(esn, FILE_LINE_MAX_SIZE, fp_in); // Entity Sequence Number
			fgets(color, FILE_LINE_MAX_SIZE, fp_in);
			fgets(line_type, FILE_LINE_MAX_SIZE, fp_in);
			fgets(line_width, FILE_LINE_MAX_SIZE, fp_in);
			fgets(display, FILE_LINE_MAX_SIZE, fp_in);
			n_property = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Property Count
			for (i = 0; i < n_property; ++i)
				property[i] = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Property pointer
			c1 = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Pointer to Arc Center Point
			p1 = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Pointer to Arc Begin Point
			p2 = atoi(fgets(file_line, FILE_LINE_MAX_SIZE, fp_in)); // Pointer to Arc End Point
			fgets(end_of_entity, FILE_LINE_MAX_SIZE, fp_in); // Arc Orientation (always 0 = anticlockwise)
			fgets(end_of_entity, FILE_LINE_MAX_SIZE, fp_in); // |~
			
			angle2 = atan2(y[p2 - esn1] - y[c1 - esn1], x[p2 - esn1] - x[c1 - esn1]);
			angle1 = atan2(y[p1 - esn1] - y[c1 - esn1], x[p1 - esn1] - x[c1 - esn1]);
			
			if (angle2 > angle1)
				length = ( angle2 - angle1 ) *
						 hypot(y[p1 - esn1] - y[c1 - esn1], x[p1 - esn1] - x[c1 - esn1]);
			else
				length = ( angle2 - angle1 + 2 * M_PI) *
						 hypot(y[p1 - esn1] - y[c1 - esn1], x[p1 - esn1] - x[c1 - esn1]);
						 
			if (length <= 1) ++n_ARC_1;
			if (length <= 2) ++n_ARC_2;
			if (length <= 5) ++n_ARC_5;
			
			if (length > max_length) {
				fprintf(fp_out, "%s", "ARC\n");
				fprintf(fp_out, "%s", esn);
				fprintf(fp_out, "%s", color);
				fprintf(fp_out, "%s", line_type);
				fprintf(fp_out, "%s", line_width);
				fprintf(fp_out, "%s", display);
				fprintf(fp_out, "%d\n", n_property);
				for (i = 0; i < n_property; ++i)
					fprintf(fp_out, "%d\n", property[i]);
				fprintf(fp_out, "%d\n", c1);
				fprintf(fp_out, "%d\n", p1);
				fprintf(fp_out, "%d\n", p2);
				fprintf(fp_out, "%s\n", "0");
				fprintf(fp_out, "%s", end_of_entity);
			}
			else
				++n_ARC_deleted;
		}
		
		else
			fprintf(fp_out, "%s", file_line);
	} while (fgets(file_line, FILE_LINE_MAX_SIZE, fp_in) != NULL);
		
	fclose(fp_in);
	fclose(fp_out);
	
	fprintf(fp_log, "Biggest section #~61:%7d points\n", n_point_max);
	
	fprintf(fp_log, "\n");
	fprintf(fp_log, "LIN\n");
	fprintf(fp_log, "Total      :%7d\n", n_LIN);
	fprintf(fp_log, "Length <= 1:%7d\n", n_LIN_1);
	fprintf(fp_log, "Length <= 2:%7d\n", n_LIN_2);
	fprintf(fp_log, "Length <= 5:%7d\n", n_LIN_5);
	fprintf(fp_log, "Deleted    :%7d\n", n_LIN_deleted);
	fprintf(fp_log, "\n");
	fprintf(fp_log, "CIR\n");
	fprintf(fp_log, "Total        :%7d\n", n_CIR);
	fprintf(fp_log, "diameter <= 1:%7d\n", n_CIR_1);
	fprintf(fp_log, "diameter <= 2:%7d\n", n_CIR_2);
	fprintf(fp_log, "diameter <= 5:%7d\n", n_CIR_5);
	fprintf(fp_log, "Deleted      :%7d\n", n_CIR_deleted);
	fprintf(fp_log, "\n");
	fprintf(fp_log, "ARC\n");
	fprintf(fp_log, "Total      :%7d\n", n_ARC);
	fprintf(fp_log, "length <= 1:%7d\n", n_ARC_1);
	fprintf(fp_log, "length <= 2:%7d\n", n_ARC_2);
	fprintf(fp_log, "length <= 5:%7d\n", n_ARC_5);
	fprintf(fp_log, "Deleted    :%7d\n", n_ARC_deleted);

}
